from russian_protowhat.Feedback import Feedback
from russian_protowhat.Test import TestFail, Fail
from functools import partial

from russian_protowhat.utils import legacy_signature


def multi(state, *tests):
    """Run multiple subtests. Return original state (for chaining).

    This function could be thought as an AND statement, since all tests it runs must pass

    Args:
        state: State instance describing student and solution code,  can be omitted if used with Ex()
        tests: one or more sub-SCTs to run.

    :Example:
        The SCT below checks two has_code cases.. ::

            Ex().multi(has_code('SELECT'), has_code('WHERE'))

        The SCT below uses ``multi`` to 'branch out' to check that
        the SELECT statement has both a WHERE and LIMIT clause.. ::

            Ex().check_node('SelectStmt', 0).multi(
                check_edge('where_clause'),
                check_edge('limit_clause')
            )
    """
    for test in iter_tests(tests):
        # assume test is function needing a state argument
        # partial state so reporter can test
        state.do_test(partial(test, state))

    # return original state, so can be chained
    return state


@legacy_signature(incorrect_msg="msg")
def check_not(state, *tests, msg):
    """Run multiple subtests that should fail. If all subtests fail, returns original state (for chaining)

    - This function is currently only tested in working with ``has_code()`` in the subtests.
    - This function can be thought as a ``NOT(x OR y OR ...)`` statement, since all tests it runs must fail
    - This function can be considered a direct counterpart of multi.

    Args:
        state: State instance describing student and solution code, can be omitted if used with Ex()
        *tests: one or more sub-SCTs to run
        msg: feedback message that is shown in case not all tests specified in ``*tests`` fail.

    :Example:

        Thh SCT below runs two has_code cases.. ::

            Ex().check_not(
                has_code('INNER'),
                has_code('OUTER'),
                incorrect_msg="Don't use `INNER` or `OUTER`!"
            )

        If students use ``INNER (JOIN)`` or ``OUTER (JOIN)`` in their code, this test will fail.

    """
    for test in iter_tests(tests):
        try:
            test(state)
        except TestFail:
            # it fails, as expected, off to next one
            continue
        return state.report(msg)

    # return original state, so can be chained
    return state


def check_or(state, *tests):
    """Test whether at least one SCT passes.

    If all of the tests fail, the feedback of the first test will be presented to the student.

    Args:
        state: State instance describing student and solution code, can be omitted if used with Ex()
        tests: one or more sub-SCTs to run

    :Example:
        The SCT below tests that the student typed either 'SELECT' or 'WHERE' (or both).. ::

            Ex().check_or(
                has_code('SELECT'),
                has_code('WHERE')
            )

        The SCT below checks that a SELECT statement has at least a WHERE c or LIMIT clause.. ::

            Ex().check_node('SelectStmt', 0).check_or(
                check_edge('where_clause'),
                check_edge('limit_clause')
            )
    """
    success = False
    first_feedback = None

    for test in iter_tests(tests):
        try:
            multi(state, test)
            success = True
        except TestFail as e:
            if not first_feedback:
                first_feedback = e.feedback
        if success:
            return state  # todo: add test

    state.do_test(Fail(first_feedback))


def check_correct(state, check, diagnose):
    """Allows feedback from a diagnostic SCT, only if a check SCT fails.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        check: An sct chain that must succeed.
        diagnose: An sct chain to run if the check fails.

    :Example:
        The SCT below tests whether students query result is correct, before running diagnostic SCTs.. ::

            Ex().check_correct(
                check_result(),
                check_node('SelectStmt')
            )

    """
    feedback = None
    try:
        multi(state, check)
    except TestFail as e:
        feedback = e.feedback

    if feedback is not None or getattr(state, "force_diagnose", False):
        try:
            multi(state, diagnose)
        except TestFail as e:
            feedback = e.feedback

    if feedback is not None:
        state.do_test(Fail(feedback))

    return state  # todo: add test


def iter_tests(tests):
    for arg in tests:
        if arg is None:
            continue

        # when input is a single test, make iterable
        if callable(arg):
            arg = [arg]

        for test in arg:
            yield test


def disable_highlighting(state):
    """Disable highlighting in the remainder of the SCT chain.

    Include this function if you want to avoid that pythonwhat marks which part of the student submission is incorrect.
    """
    return state.to_child(highlighting_disabled=True)


def fail(state, msg="fail"):
    """Always fails the SCT, with an optional msg.

    This function takes a single argument, ``msg``, that is the feedback given to the student.
    Note that this would be a terrible idea for grading submissions, but may be useful while writing SCTs.
    For example, failing a test will highlight the code as if the previous test/check had failed.
    """
    _msg = state.build_message(msg)
    state.report(_msg)

    return state



from typing import TypeVar, Generic, Union, List, Dict, Tuple
from collections import Mapping
from ast import NodeVisitor
import inspect
import importlib

from russian_protowhat.utils_messaging import get_ord


class Selector(NodeVisitor):
    def __init__(self, target_cls, target_cls_name=None, strict=True, priority=None, include_head=False):
        self.target_cls = target_cls
        self.target_cls_name = target_cls_name
        self.strict = strict
        self.priority = priority if priority else self._get_node_priority(target_cls)
        self.out = []
        self.include_head = include_head

    def visit(self, node, head=False):
        """
        Find child nodes at the first level
        and keep searching if their children have a lower priority.
        If self.include_head is True then the starting node will also be considered.
        """
        if (not head or self.include_head) and self.is_match(node):
            self.out.append(node)
        if self.has_priority_over(node) or head:
            return super().visit(node)

    def visit_list(self, lst):
        # this allows the root to be a list
        for item in lst:
            self.visit(item)

    def is_match(self, node):
        if self.strict:
            if type(node) is self.target_cls:
                return True
            else:
                return False
        else:
            if isinstance(node, self.target_cls) and (
                self.target_cls_name is None
                or self.target_cls_name == node.__class__.__name__
            ):
                return True
            else:
                return False

    def has_priority_over(self, node):
        return self.priority > self._get_node_priority(node)

    def _get_node_priority(self, node):
        return getattr(node, "_priority", 0)


T = TypeVar("T")


class DispatcherInterface(Generic[T]):
    def find(self, name: str, node: T, *args, **kwargs) -> Union[List[T], Dict[str, T]]:
        # todo: document signature, strategy kwarg (depth/breadth first)
        raise NotImplementedError

    def select(self, path: Union[str, Tuple], node: T) -> Union[T, List[T]]:
        raise NotImplementedError

    @staticmethod
    def _path_str_to_list(path):
        steps = path.split(".")

        def parse_int(x):
            try:
                return int(x)
            except ValueError:
                return x

        return [parse_int(step) for step in steps if step]

    def parse(self, code: str):
        raise NotImplementedError


class Dispatcher(DispatcherInterface):
    def __init__(self, node_cls, nodes=None, ast_mod=None, safe_parsing=True):
        """Wrapper to instantiate and use a Selector using node names."""
        self.node_cls = node_cls
        self.nodes = nodes or getattr(ast_mod, "nodes", {})
        self.ast_mod = ast_mod
        self.safe_parsing = safe_parsing

        self.ParseError = getattr(
            self.ast_mod, "ParseError", type("ParseError", (Exception,), {})
        )

    def find(self, name, node, *args, **kwargs):
        if self.nodes and name in self.nodes:
            ast_cls = self.nodes[name]
            strict_selector = True
        else:
            ast_cls = self.node_cls
            strict_selector = False

        selector = Selector(
            ast_cls, target_cls_name=name, strict=strict_selector, *args, **kwargs
        )
        selector.visit(node, head=True)

        return selector.out

    def select(self, spec, node):
        result = node
        if isinstance(spec, tuple):
            raise ValueError(
                "This dispatcher currently doesn't support tuple specs for select"
            )
        if isinstance(spec, str):
            spec = self._path_str_to_list(spec)
        for step in spec:
            if isinstance(step, str):
                if isinstance(result, Mapping):
                    result = result.get(step, None)
                else:
                    result = getattr(result, step, None)
            elif isinstance(step, int):
                result = result[step] if len(result) > step else None
            if result is None:
                break
        return result

    def parse(self, code):
        try:
            return self.ast_mod.parse(code, strict=True)
        except self.ParseError as e:
            if self.safe_parsing:
                return e
            else:
                raise e

    def describe(self, node, msg, field="", **kwargs):
        speaker = getattr(self.ast_mod, "speaker", None)

        if kwargs.get("index") is not None:
            phrase = "{} entry in the " if field else "{} "
            kwargs["index"] = phrase.format(get_ord(kwargs["index"] + 1))
        else:
            kwargs["index"] = ""

        if speaker:
            return self.ast_mod.speaker.describe(node, field=field, fmt=msg, **kwargs)

    @classmethod
    def from_module(cls, mod):
        if isinstance(mod, str):
            mod = importlib.import_module(mod)

        ast_nodes = getattr(mod, "nodes", None)
        if ast_nodes is None:
            ast_nodes = {
                k: v
                for k, v in vars(mod).items()
                if (inspect.isclass(v) and issubclass(v, mod.AstNode))
            }
        dispatcher = cls(mod.AstNode, nodes=ast_nodes, ast_mod=mod)
        return dispatcher



from pathlib import Path


def check_file(
    state,
    path,
    missing_msg="Вы создали файл `{}`?",
    is_dir_msg="Хотите проверить файл `{}`, но нашли каталог.",
    parse=True,
    solution_code=None,
):
    """Test whether file exists, and make its contents the student code.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        path: expected location of the file
        missing_msg: feedback message if no file is found in the expected location
        is_dir_msg: feedback message if the path is a directory instead of a file
        parse: If ``True`` (the default) the content of the file is interpreted as code in the main exercise technology.
            This enables more checks on the content of the file.
        solution_code: this argument can be used to pass the expected code for the file
            so it can be used by subsequent checks.

    Note:
        This SCT fails if the file is a directory.

    :Example:

        To check if a user created the file ``my_output.txt`` in the subdirectory ``resources``
        of the directory where the exercise is run, use this SCT::

            Ex().check_file("resources/my_output.txt", parse=False)
    """

    path_obj = Path(path)
    if not path_obj.exists():
        state.report(missing_msg.format(path))  # test file exists
    if path_obj.is_dir():
        state.report(is_dir_msg.format(path))  # test its not a dir

    code = get_file_content(path_obj)

    sol_kwargs = {"solution_code": solution_code, "solution_ast": None}
    if solution_code:
        sol_kwargs["solution_ast"] = (
            state.parse(solution_code, test=False) if parse else False
        )

    child_state = state.to_child(
        append_message="Мы проверили файл `{}`. ".format(path),
        student_code=code,
        student_ast=state.parse(code) if parse else False,
        **sol_kwargs
    )

    child_state.path = path_obj  # .parent + .name

    return child_state


def has_dir(state, path, msg="Вы создали каталог `{}`?"):
    """Test whether a directory exists.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        path: expected location of the directory
        msg: feedback message if no directory is found in the expected location

    :Example:

        To check if a user created the subdirectory ``resources``
        in the directory where the exercise is run, use this SCT::

            Ex().has_dir("resources")
    """
    if not Path(path).is_dir():
        state.report(msg.format(path))

    return state


# helper functions
def load_file(relative_path, prefix=""):
    # the prefix can be partialed
    # so it's not needed to copy the common part of paths
    path = Path(prefix, relative_path)

    return get_file_content(path)


def get_file_content(path):
    if not isinstance(path, Path):
        path = Path(path)

    try:
        content = path.read_text(encoding="utf-8")
    except:
        content = None

    return content



import setuptools


setuptools.setup(
    # Here is the module name.
    name="russian_protowhat",

    # version of the module
    version="0.0.1",

    # Name of Author
    author="nikitqa",

    # your Email address
    author_email="investinlipetsk@gmail.com",

    # #Small Description about module
    # description="adding number",

    # long_description=long_description,

    # Specifying that we are using markdown file for description
    # long_description=long_description,
    long_description_content_type="text/markdown",

    # if module has dependencies i.e. if your package rely on other package at pypi.org
    # then you must add there, in order to download every requirement of package

    #     install_requires=[
    #      "package1",
    #    "package2",
    #    ],
    python_requires='>=3.9',

    license="MIT",

    # classifiers like program is suitable for python3, just leave as it is.
    classifiers=[
        "Programming Language :: Python :: 3.9.0",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import re

from functools import partial, wraps

from russian_protowhat.Feedback import Feedback

MSG_CHECK_FALLBACK = "Ваше представление неверно. Попробуйте еще раз!"


def requires_ast(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        state = kwargs.get("state", args[0] if len(args) else None)
        state_ast = [state.student_ast, state.solution_ast]

        # fail if no ast parser in use
        if any(ast is None for ast in state_ast):
            raise TypeError(
                "Пытаюсь использовать ast, но это None. Вы используете парсер? {} {}".format(
                    args, kwargs
                )
            )

        # check whether the parser passed or failed for some code
        # if safe_parsing is enabled in the Dispatcher (otherwise an exception would be raised earlier)
        ParseError = state.ast_dispatcher.ParseError

        parse_fail = any(isinstance(ast, ParseError) for ast in state_ast)

        if parse_fail:
            return state  # skip test
        else:
            return f(*args, **kwargs)  # proceed with test

    return wrapper


@requires_ast
def check_node(
    state,
    name,
    index=0,
    missing_msg="Проверьте {ast_path}. Не удалось найти {index}{node_name}.",
    priority=None,
):
    """Select a node from abstract syntax tree (AST), using its name and index position.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        name : the name of the abstract syntax tree node to find.
        index: the position of that node (see below for details).
        missing_msg: feedback message if node is not in student AST.
        priority: the priority level of the node being searched for. This determines whether to
                  descend into other AST nodes during the search. Higher priority nodes descend
                  into lower priority. Currently, the only important part of priority is that
                  setting a very high priority (e.g. 99) will search every node.


    :Example:
        If both the student and solution code are.. ::

            SELECT a FROM b; SELECT x FROM y;

        then we can focus on the first select with::

            # approach 1: with manually created State instance
            state = State(*args, **kwargs)
            new_state = check_node(state, 'SelectStmt', 0)

            # approach 2:  with Ex and chaining
            new_state = Ex().check_node('SelectStmt', 0)

    """
    df = partial(state.ast_dispatcher.find, name, priority=priority)

    sol_stmt_list = df(state.solution_ast)
    try:
        sol_stmt = sol_stmt_list[index]
    except IndexError:
        raise IndexError("Не удается получить оператор %s по индексу %s" % (name, index))

    stu_stmt_list = df(state.student_ast)
    try:
        stu_stmt = stu_stmt_list[index]
    except IndexError:
        # use speaker on ast dialect module to get message, or fall back to generic
        ast_path = state.get_ast_path() or "highlighted code"
        _msg = state.ast_dispatcher.describe(
            sol_stmt, missing_msg, index=index, ast_path=ast_path
        )
        if _msg is None:
            _msg = MSG_CHECK_FALLBACK
        state.report(_msg)

    return state.to_child(student_ast=stu_stmt, solution_ast=sol_stmt)


@requires_ast
def check_edge(
    state,
    name,
    index=0,
    missing_msg="Проверьте {ast_path}. Не удалось найти {index}{field_name}.",
):
    """Select an attribute from an abstract syntax tree (AST) node, using the attribute name.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        name: the name of the attribute to select from current AST node.
        index: entry to get from a list field. If too few entires, will fail with missing_msg.
        missing_msg: feedback message if attribute is not in student AST.

    :Example:
        If both the student and solution code are.. ::

            SELECT a FROM b; SELECT x FROM y;

        then we can get the from_clause using ::

            # approach 1: with manually created State instance -----
            state = State(*args, **kwargs)
            select = check_node(state, 'SelectStmt', 0)
            clause = check_edge(select, 'from_clause')

            # approach 2: with Ex and chaining ---------------------
            select = Ex().check_node('SelectStmt', 0)           # get first select statement
            clause =  select.check_edge('from_clause', None)    # get from_clause (a list)
            clause2 = select.check_edge('from_clause', 0)       # get first entry in from_clause
    """

    def select(node_name, node):
        attr = state.ast_dispatcher.select(node_name, node)
        if attr and isinstance(attr, list) and index is not None:
            attr = attr[index]
        return attr

    try:
        sol_attr = select(name, state.solution_ast)
    except IndexError:
        raise IndexError("Не удается получить атрибут %s" % name)

    # use speaker on ast dialect module to get message, or fall back to generic
    ast_path = state.get_ast_path() or "highlighted code"
    _msg = state.ast_dispatcher.describe(
        state.student_ast, missing_msg, field=name, index=index, ast_path=ast_path
    )
    if _msg is None:
        _msg = MSG_CHECK_FALLBACK

    try:
        stu_attr = select(name, state.student_ast)
    except:
        state.report(_msg)

    # fail if attribute exists, but is none only for student
    if stu_attr is None and sol_attr is not None:
        state.report(_msg)

    return state.to_child(student_ast=stu_attr, solution_ast=sol_attr)


def has_code(
    state,
    text,
    incorrect_msg="Проверьте {ast_path}. Программа проверки ожидала найти {text}.",
    fixed=False,
):
    """Test whether the student code contains text.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        text : text that student code must contain. Can be a regex pattern or a simple string.
        incorrect_msg: feedback message if text is not in student code.
        fixed: whether to match text exactly, rather than using regular expressions.

    Note:
        Functions like ``check_node`` focus on certain parts of code.
        Using these functions followed by ``has_code`` will only look
        in the code being focused on.

    :Example:
        If the student code is.. ::

            SELECT a FROM b WHERE id < 100

        Then the first test below would (unfortunately) pass, but the second would fail..::

            # contained in student code
            Ex().has_code(text="id < 10")

            # the $ means that you are matching the end of a line
            Ex().has_code(text="id < 10$")

        By setting ``fixed = True``, you can search for fixed strings::

            # without fixed = True, '*' matches any character
            Ex().has_code(text="SELECT * FROM b")               # passes
            Ex().has_code(text="SELECT \\\\* FROM b")             # fails
            Ex().has_code(text="SELECT * FROM b", fixed=True)   # fails

        You can check only the code corresponding to the WHERE clause, using ::

            where = Ex().check_node('SelectStmt', 0).check_edge('where_clause')
            where.has_code(text = "id < 10)

    """
    stu_ast = state.student_ast
    stu_code = state.student_code

    # fallback on using complete student code if no ast
    ParseError = state.ast_dispatcher.ParseError

    def get_text(ast, code):
        if isinstance(ast, ParseError):
            return code
        try:
            return ast.get_text(code) or ""
        except:
            return code

    stu_text = get_text(stu_ast, stu_code)

    _msg = incorrect_msg.format(
        ast_path=state.get_ast_path() or "highlighted code", text=text
    )

    # either simple text matching or regex test
    res = text in stu_text if fixed else re.search(text, stu_text)

    if not res:
        state.report(_msg)

    return state


@requires_ast
def has_equal_ast(
    state,
    incorrect_msg="Проверьте {ast_path}. {extra}",
    sql=None,
    start=["expression", "subquery", "sql_script"][0],
    exact=None,
):
    """Test whether the student and solution code have identical AST representations

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        incorrect_msg: feedback message if student and solution ASTs don't match
        sql  : optional code to use instead of the solution ast that is zoomed in on.
        start: if ``sql`` arg is used, the parser rule to parse the sql code.
               One of 'expression' (the default), 'subquery', or 'sql_script'.
        exact: whether to require an exact match (True), or only that the
               student AST contains the solution AST. If not specified, this
               defaults to ``True`` if ``sql`` is not specified, and to ``False``
               if ``sql`` is specified. You can always specify it manually.

    :Example:

        Example 1 - Suppose the solution code is ::

            SELECT * FROM cities

        and you want to verify whether the `FROM` part is correct: ::

            Ex().check_node('SelectStmt').from_clause().has_equal_ast()

        Example 2 - Suppose the solution code is ::

            SELECT * FROM b WHERE id > 1 AND name = 'filip'

        Then the following SCT makes sure ``id > 1`` was used somewhere in the WHERE clause.::

            Ex().check_node('SelectStmt') \\/
                .check_edge('where_clause') \\/
                .has_equal_ast(sql = 'id > 1')

    """
    ast = state.ast_dispatcher.ast_mod
    sol_ast = state.solution_ast if sql is None else ast.parse(sql, start)

    # if sql is set, exact defaults to False.
    # if sql not set, exact defaults to True.
    if exact is None:
        exact = sql is None

    stu_rep = repr(state.student_ast)
    sol_rep = repr(sol_ast)

    def get_str(ast, code, sql):
        if sql:
            return sql
        if isinstance(ast, str):
            return ast
        try:
            return ast.get_text(code)
        except:
            return None

    sol_str = get_str(state.solution_ast, state.solution_code, sql)
    _msg = incorrect_msg.format(
        ast_path=state.get_ast_path()
        or ("code" if state.highlighting_disabled else "highlighted code"),
        extra="Программа проверки ожидала найти там `{}`.".format(sol_str)
        if sol_str
        else "Что-то не так",
    )
    if (exact and (sol_rep != stu_rep)) or (not exact and (sol_rep not in stu_rep)):
        state.report(_msg)

    return state


def has_parsed_ast(state):
    asts = [state.student_ast, state.solution_ast]
    if any(isinstance(c, state.ast_dispatcher.ParseError) for c in asts):
        state.report("AST did not parse")

    return state



def get_ord(num):
    assert num > 0, "используйте строго положительные числа в get_ord()"
    nums = {
        1: "первый",
        2: "второй",
        3: "третий",
        4: "четвертый",
        5: "пятый",
        6: "шестой",
        7: "седьмой",
        8: "восьмой",
        9: "девятый",
        10: "десятый",
    }
    if num in nums:
        return nums[num]
    else:
        return (
            {1: "{}й", 2: "{}й", 3: "{}й"}.get(
                num if (num < 20) else (num % 10), "{}й"
            )
        ).format(num)


def get_times(num):
    nums = {1: "однажды", 2: "дважды"}
    if num in nums:
        return nums[num]
    else:
        return "%s раз" % get_num(num)


def get_num(num):
    nums = {
        0: "нет",
        1: "один",
        2: "два",
        3: "три",
        4: "четыре",
        5: "пять",
        6: "шесть",
        7: "семь",
        8: "восемь",
    }
    if num in nums:
        return nums[num]
    else:
        return str(num)



from ast import AST
from collections import OrderedDict


class DumpConfig:
    def __init__(
        self,
        is_node=lambda node: isinstance(node, AST),
        node_type=lambda node: node.__class__.__name__,
        fields_iter=lambda node: node._fields,
        field_val=lambda node, field: getattr(node, field, None),
        is_list=lambda node: isinstance(node, list),
        list_iter=id,
        leaf_val=id,
    ):
        """
        Configuration to convert a node tree to the dump format

        The default configuration can be used to dump a tree of AstNodes
        """
        self.is_node = is_node
        self.node_type = node_type
        self.fields_iter = fields_iter
        self.field_val = field_val
        self.is_list = is_list
        self.list_iter = list_iter
        self.leaf_val = leaf_val


def dump(node, config):
    """
    Convert a node tree to a simple nested dict

    All steps in this conversion are configurable using DumpConfig

    dump dictionary node: {"type": str, "data": dict}
    """
    if config.is_node(node):
        fields = OrderedDict()
        for name in config.fields_iter(node):
            attr = config.field_val(node, name)
            if attr is not None:
                fields[name] = dump(attr, config)
        return {"type": config.node_type(node), "data": fields}
    elif config.is_list(node):
        return [dump(x, config) for x in config.list_iter(node)]
    else:
        return config.leaf_val(node)


class AstNode(AST):
    _fields = ()
    _priority = 1

    def get_text(self, full_text=None):
        raise NotImplementedError()

    def get_position(self):
        raise NotImplementedError()

    def __str__(self):
        els = [k for k in self._fields if getattr(self, k, None) is not None]
        return "{}: {}".format(self.__class__.__name__, ", ".join(els))

    def __repr__(self):
        field_reps = [
            (k, repr(getattr(self, k)))
            for k in self._fields
            if getattr(self, k, None) is not None
        ]
        args = ", ".join("{} = {}".format(k, v) for k, v in field_reps)
        return "{}({})".format(self.__class__.__name__, args)


class ParseError(Exception):
    pass


class AstModule:
    """Subclasses can be used to instantiate a Dispatcher"""

    AstNode = AstNode
    ParseError = ParseError
    nodes = dict()
    speaker = None

    @classmethod
    def parse(cls, code, **kwargs):
        raise NotImplementedError("This method needs to be defined in a subclass.")

    @classmethod
    def dump(cls, tree):
        return dump(tree, DumpConfig())

    # methods below are for updating an AstModule subclass based on data in the dump dictionary format --

    @classmethod
    def load(cls, node):
        if not isinstance(node, dict):
            return node  # return primitives

        type_str = node["type"]
        data = node["data"]
        obj = cls._instantiate_node(type_str, tuple(data.keys()))
        for field_name, value in data.items():
            if isinstance(value, (list, tuple)):
                child = [cls.load(entry) for entry in value]
            else:
                child = cls.load(value)
            setattr(obj, field_name, child)

        return obj

    @classmethod
    def _instantiate_node(cls, type_str, fields):
        # TODO: implement on AstNode (+ interface to get classes)
        node_cls = cls.nodes.get(type_str, None)
        if not node_cls:
            node_cls = type(type_str, (cls.AstNode,), {"_fields": fields})
            cls.nodes[type_str] = node_cls

        return node_cls()



from russian_protowhat.Feedback import Feedback


class TestFail(Exception):
    def __init__(self, feedback, payload):
        super().__init__(feedback.message)
        self.feedback = feedback
        self.payload = payload


class Test:
    """
    The basic Test. It should only contain a failure message, as all tests should result in
    a failure message when they fail.

    Note:
        This test should not be used by itself, subclasses should be used.

    Attributes:
        feedback (str): A string containing the failure message in case the test fails.
        result (bool): True if the test succeed, False if it failed. None if it hasn't been tested yet.
    """

    def __init__(self, feedback):
        """
        Initialize the standard test.

        Args:
            feedback: string or Feedback object
        """
        if issubclass(type(feedback), Feedback):
            self.feedback = feedback
        elif issubclass(type(feedback), str):
            self.feedback = Feedback(feedback)
        else:
            raise TypeError(
                "When creating a test, specify either a string or a Feedback object"
            )

        self.result = None

    def __call__(self):
        """
        Wrapper around specific tests. Tests only get one chance.
        """
        if self.result is None:
            try:
                self.test()
                self.result = bool(self.result)
            except:
                self.result = False

    def test(self):
        """
        Perform the actual test.
        """
        raise NotImplementedError

    def get_feedback(self):
        return self.feedback

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, repr(vars(self)))


class Fail(Test):
    def test(self):
        self.result = False





def has_chosen(state, correct, msgs):
    """Verify exercises of the type MultipleChoiceExercise

    Args:
        state:    State instance describing student and solution code. Can be omitted if used with Ex().
        correct:  index of correct option, where 1 is the first option.
        msgs  :    list of feedback messages corresponding to each option.

    :Example:
        The following SCT is for a multiple choice exercise with 2 options, the first
        of which is correct.::

            Ex().has_chosen(1, ['Correct!', 'Incorrect. Try again!'])
    """

    ctxt = {}
    exec(state.student_code, globals(), ctxt)
    sel_indx = ctxt["selected_option"]
    if sel_indx != correct:
        state.report(msgs[sel_indx - 1])
    else:
        state.reporter.success_msg = msgs[correct - 1]

    return state


def allow_errors(state):
    """
    Allow running the student code to generate errors.

    This has to be used only once for every time code is executed or a different xwhat library is used.
    In most exercises that means it should be used just once.

    :Example:
        The following SCT allows the student code to generate errors::

            Ex().allow_errors()
    """
    state.reporter.allow_errors()

    return state


def success_msg(state, msg):
    """
    Changes the success message to display if submission passes.

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        msg  : feedback message if student and solution ASTs don't match

    :Example:
        The following SCT changes the success message::

            Ex().success_msg("You did it!")

    """
    state.reporter.success_msg = msg

    return state



from functools import wraps


def _debug(state, msg="", on_error=False):
    """
    This SCT function makes the SCT fail with a message containing debugging information
    and highlights the focus of the SCT at that point.
    """
    check_history = [
        s.creator["type"] for s in state.state_history if getattr(s, "creator", None)
    ]

    feedback = ""
    if msg:
        feedback += msg + "\n"

    if check_history:
        feedback += "SCT function state history: `{}`".format(" > ".join(check_history))

    if state.reporter.tests:
        feedback += "\nLast test: `{}`".format(repr(state.reporter.tests[-1]))

    if not on_error:
        # latest highlight added automatically
        state.report(feedback)
    else:
        # debug on next failure
        state.debug = True
        # or at the end (to prevent debug mode in production)
        state.reporter.fail = True

    return state


def legacy_signature(**kwargs_mapping):
    """
    This decorator makes it possible to call a function using old argument names
    when they are passed as keyword arguments.

    @legacy_signature(old_arg1='arg1', old_arg2='arg2')
    def func(arg1, arg2=1):
        return arg1 + arg2

    func(old_arg1=1) == 2
    func(old_arg1=1, old_arg2=2) == 3
    """

    def signature_decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            redirected_kwargs = {
                kwargs_mapping[k] if k in kwargs_mapping else k: v
                for k, v in kwargs.items()
            }
            return f(*args, **redirected_kwargs)

        return wrapper

    return signature_decorator



import re
from collections import Counter

import markdown2

from russian_protowhat.Feedback import Feedback
from russian_protowhat.Test import Test

"""
This file holds the reporter class.
"""


class TestRunner:
    def __init__(self):
        self.tests = []

    def do_test(self, test):
        """Raise failing test.

        Raise a ``TestFail`` object, containing the feedback message and highlight information.
        """
        result = None
        feedback = None
        test()
        if isinstance(test, Test):
            self.tests.append(test)
            result = test.result
            if not result:
                feedback = test.get_feedback()

        return result, feedback

    def do_tests(self, tests):
        return [self.do_test(test) for test in tests]

    @property
    def failures(self):
        return list(filter(lambda test: test.result is False, self.tests))

    @property
    def has_failed(self):
        return len(self.failures) > 0


class TestRunnerProxy(TestRunner):
    # Checks using this might be split into more atomic checks
    def __init__(self, runner: TestRunner):
        super().__init__()
        self.runner = runner

    def do_test(self, test):
        if isinstance(test, Test):
            self.tests.append(test)
        return self.runner.do_test(test)


class Reporter(TestRunnerProxy):
    """Do reporting.

    This class holds the feedback- or success message and tracks whether there are failed tests
    or not. All tests are executed trough do_test() in the Reporter.
    """

    # This is the offset for ANTLR
    ast_highlight_offset = {"column_start": 1, "column_end": 1}

    def __init__(self, runner=None, errors=None, highlight_offset=None):
        super().__init__(runner or TestRunner())
        self.fail = False
        self.errors = errors
        self.errors_allowed = False
        self.highlight_offset = highlight_offset or {}
        self.success_msg = "Great work!"

    def get_errors(self):
        return self.errors

    def allow_errors(self):
        self.errors_allowed = True

    def build_failed_payload(self, feedback: Feedback):
        highlight = Counter()
        code_highlight = feedback.get_highlight_data()

        path = code_highlight.get("path", None)
        if path is not None:
            del code_highlight["path"]

        if code_highlight:
            highlight.update(self.highlight_offset)
            if "line_start" in highlight and "line_end" not in highlight:
                highlight["line_end"] = highlight["line_start"]

            highlight.update(code_highlight)
            highlight.update(self.ast_highlight_offset)

        if path is not None:
            highlight["path"] = str(path)

        return {
            "correct": False,
            "message": Reporter.to_html(feedback.message),
            **highlight,
        }

    def build_final_payload(self):
        if self.fail or self.errors and not self.errors_allowed:
            feedback_msg = "Ваш код вызвал ошибку. Исправьте и попробуйте еще раз!"
            return {"correct": False, "message": Reporter.to_html(feedback_msg)}
        else:
            return {"correct": True, "message": Reporter.to_html(self.success_msg)}

    @staticmethod
    def to_html(msg):
        return re.sub("<p>(.*)</p>", "\\1", markdown2.markdown(msg)).strip()



import inspect
import copy
import builtins
from functools import wraps, reduce, partial


def state_dec_gen(state_cls, attr_scts):
    def state_dec(f):
        """Decorate check_* functions to return F chain if no state passed"""

        @wraps(f)
        def wrapper(*args, **kwargs):
            state = kwargs.get("state", args[0] if len(args) else None)
            if isinstance(state, state_cls):
                return f(*args, **kwargs)
            else:
                return F._from_func(f, *args, _attr_scts=attr_scts, **kwargs)

        return wrapper

    return state_dec


def link_to_state(check):
    @wraps(check)
    def wrapper(state, *args, **kwargs):
        new_state = check(state, *args, **kwargs)
        if (
            new_state != state
            and hasattr(new_state, "creator")
            and not isinstance(check, F)
        ):
            ba = inspect.signature(check).bind(state, *args, **kwargs)
            ba.apply_defaults()
            new_state.creator = {"type": check.__name__, "args": ba.arguments}

        return new_state

    return wrapper


class Chain:
    def __init__(self, state, attr_scts=None):
        self._state = state
        self._crnt_sct = None
        self._waiting_on_call = False
        self._attr_scts = {} if attr_scts is None else attr_scts

    def _double_attr_error(self):
        raise AttributeError(
            "Did you forget to call a statement? "
            "e.g. Ex().check_list_comp.check_body()"
        )

    def __getattr__(self, attr):
        # Enable fast attribute access
        if attr == "_attr_scts":
            raise AttributeError("Prevent getattr recursion on copy")
        attr_scts = self._attr_scts
        if attr not in attr_scts:
            raise AttributeError("No SCT named %s" % attr)
        elif self._waiting_on_call:
            self._double_attr_error()
        else:
            # make a copy to return,
            # in case someone does: chain = chain_of_checks; chain.a; chain.b
            return self._sct_copy(attr_scts[attr])

    def __call__(self, *args, **kwargs):
        self._state = link_to_state(self._crnt_sct)(self._state, *args, **kwargs)
        self._waiting_on_call = False
        return self

    def __rshift__(self, f):
        if self._waiting_on_call:
            self._double_attr_error()
        elif isinstance(f, Chain) and not isinstance(f, F):
            raise BaseException(
                "did you use a result of the Ex() function on the right hand side of the >> operator?"
            )
        elif not callable(f):
            raise BaseException(
                "right hand side of >> operator should be an SCT, so must be callable!"
            )
        else:
            chain = self._sct_copy(f)
            return chain()

    def _sct_copy(self, f):
        chain = copy.copy(self)
        chain._crnt_sct = f
        chain._waiting_on_call = True
        return chain


class F(Chain):
    def __init__(self, stack=None, attr_scts=None):
        super().__init__(None, attr_scts)
        self._stack = [] if stack is None else stack

    def __call__(self, *args, **kwargs):
        if self._crnt_sct:
            # calling an SCT function (after attribute access)
            call_data = (self._crnt_sct, args, kwargs)
            return self.__class__(self._stack + [call_data], self._attr_scts)
        else:
            # running the chain
            state = kwargs.get("state") or args[0]
            return reduce(
                lambda s, cd: self._call_from_data(s, *cd), self._stack, state
            )

    @staticmethod
    def _call_from_data(state, f, args, kwargs):
        return link_to_state(f)(state, *args, **kwargs)

    @classmethod
    def _from_func(cls, f, *args, _attr_scts=None, **kwargs):
        """Creates a function chain starting with the specified SCT (f), and its arguments."""
        func_chain = cls(attr_scts=_attr_scts)
        func_chain._stack.append([f, args, kwargs])
        return func_chain


class ExGen:
    def __init__(self, root_state, attr_scts):
        self.root_state = root_state
        self.attr_scts = attr_scts

    def __call__(self, state=None):
        """Returns the current code state as a Chain instance.

        This allows SCTs to be run without including their 1st argument, ``state``.

        When writing SCTs on DataCamp, no State argument to ``Ex`` is necessary.
        The exercise State is built for you.

        Args:
            state: a State instance, which contains the student/solution code and results.

        :Example:

            code ::

                # How to write SCT on DataCamp.com
                Ex().has_code(text="SELECT id")

                # Experiment locally - chain off of Ex(), created from state
                state = SomeStateProducingFunction()
                Ex(state).has_code(text="SELECT id")

                # Experiment locally - no Ex(), create and pass state explicitly
                state = SomeStateProducingFunction()
                has_code(state, text="SELECT id")

        """
        if state is None and self.root_state is None:
            raise Exception("explicitly pass state to Ex, or set Ex.root_state")

        return Chain(state or self.root_state, attr_scts=self.attr_scts)


Ex = ExGen(None, {})


def get_checks_dict(checks_module):
    return {
        k: v
        for k, v in vars(checks_module).items()
        if k not in builtins.__dict__
        if not k.startswith("__")
        if callable(v)
    }


def create_sct_context(State, sct_dict, root_state=None):
    state_dec = state_dec_gen(State, sct_dict)
    sct_ctx = {k: state_dec(v) for k, v in sct_dict.items()}

    ctx = {
        **sct_ctx,
        "state_dec": state_dec,  # needed by ext packages
        "Ex": ExGen(root_state, sct_ctx),
        "F": partial(F, attr_scts=sct_ctx),
    }

    return ctx



import os
import re

from pathlib import Path
from russian_protowhat.Feedback import InstructorError

# env vars
BASH_HISTORY_PATH_ENV = "BASH_HISTORY_PATH"
BASH_HISTORY_INFO_PATH_ENV = "BASH_HISTORY_INFO"

# env var defaults
os.environ[BASH_HISTORY_PATH_ENV] = os.environ.get(
    BASH_HISTORY_PATH_ENV, "/home/repl/.bash_history"
)
os.environ[BASH_HISTORY_INFO_PATH_ENV] = os.environ.get(
    BASH_HISTORY_INFO_PATH_ENV, "/home/repl/.bash_history_info"
)


def update_bash_history_info(bash_history_path=None):
    """Store the current number of commands in the bash history

    ``get_bash_history`` can use this info later to get only newer commands.

    Depending on the wanted behaviour this function should be called
    at the start of the exercise or every time the exercise is submitted.

    Import using ``from russian_protowhat.checks import update_bash_history_info``.
    """
    if bash_history_path is None:
        bash_history_path = os.environ[BASH_HISTORY_PATH_ENV]
    with open(bash_history_path, encoding="utf-8") as f:
        Path(os.environ[BASH_HISTORY_INFO_PATH_ENV]).write_text(
            str(len(f.readlines())), encoding="utf-8"
        )


def get_bash_history_info():
    try:
        old_history_length = int(
            Path(os.environ[BASH_HISTORY_INFO_PATH_ENV]).read_text()
        )
    except FileNotFoundError:
        raise InstructorError(
            "`update_bash_history_info` wasn't called"
        )

    return old_history_length


def get_bash_history(full_history=False, bash_history_path=None):
    """Get the commands in the bash history

    Args:
        full_history (bool): if true, returns all commands in the bash history,
            else only return the commands executed after the last bash history info update
        bash_history_path (str | Path): path to the bash history file

    Returns:
        a list of commands (empty if the file is not found)

    Import from ``from russian_protowhat.checks import get_bash_history``.
    """
    if bash_history_path is None:
        bash_history_path = os.environ[BASH_HISTORY_PATH_ENV]
    try:
        with open(bash_history_path, encoding="utf-8") as f:
            history = f.readlines()
            old_history_length = get_bash_history_info()
            return history if full_history else history[old_history_length:]
    except FileNotFoundError:
        return []


"""
Design considerations

use case:
bash history used multiple times in SCT
solution:
separate bash history info update

use case:
bash history is not empty from earlier exercise or session (that might not use this SCT)
solution:
update info in PEC or exercise init

If updating bash history info in PEC, only use bash history in check_correct
as it only has access to the latest commands
"""


def has_command(state, pattern, msg, fixed=False, commands=None):
    """Test whether the bash history has a command matching the pattern

    Args:
        state: State instance describing student and solution code. Can be omitted if used with Ex().
        pattern: text that command must contain (can be a regex pattern or a simple string)
        msg: feedback message if no matching command is found
        fixed: whether to match text exactly, rather than using regular expressions
        commands: the bash history commands to check against.
            By default this will be all commands since the last bash history info update.
            Otherwise pass a list of commands to search through, created by calling the helper function
            ``get_bash_history()``.

    Note:
        The helper function ``update_bash_history_info(bash_history_path=None)``
        needs to be called in the pre-exercise code in exercise types that don't have
        built-in support for bash history features.

    Note:
        If the bash history info is updated every time code is submitted
        (by using ``update_bash_history_info()`` in the pre-exercise code),
        it's advised to only use this function as the second part of a ``check_correct()``
        to help students debug the command they haven't correctly run yet.
        Look at the examples to see what could go wrong.

        If bash history info is only updated at the start of an exercise,
        this can be used everywhere as the (cumulative) commands from all submissions are known.

    :Example:

        The goal of an exercise is to use ``man``.

        If the exercise doesn't have built-in support for bash history SCTs,
        update the bash history info in the pre-exercise code::

            update_bash_history_info()

        In the SCT, check whether a command with ``man`` was used::

            Ex().has_command("$man\s", "Your command should start with ``man ...``.")

    :Example:

        The goal of an exercise is to use ``touch`` to create two files.

        In the pre-exercise code, put::

            update_bash_history_info()

        This SCT can cause problems::

            Ex().has_command("touch.*file1", "Use `touch` to create `file1`")
            Ex().has_command("touch.*file2", "Use `touch` to create `file2`")

        If a student submits after running ``touch file0 && touch file1`` in the console,
        they will get feedback to create ``file2``.
        If they submit again after running ``touch file2`` in the console,
        they will get feedback to create ``file1``, since the SCT only has access
        to commands after the last bash history info update (only the second command in this case).
        Only if they execute all required commands in a single submission the SCT will pass.

        A better SCT in this situation checks the outcome first
        and checks the command to help the student achieve it::

            Ex().check_correct(
                check_file('file1', parse=False),
                has_command("touch.*file1", "Use `touch` to create `file1`")
            )
            Ex().check_correct(
                check_file('file2', parse=False),
                has_command("touch.*file2", "Use `touch` to create `file2`")
            )

    """
    if commands is None:
        commands = get_bash_history()
    if not commands:
        state.report("В поисках исполняемой команды оболочки мы ничего не нашли.")
    if not state.is_root:
        raise InstructorError(
            "`has_command()` should only be called from the root state, `Ex()`."
        )

    correct = False
    for command in commands:
        # similar to has_code
        if pattern in command if fixed else re.search(pattern, command):
            correct = True
            break

    if not correct:
        state.report(msg)

    return state


"""
# Other SCTs using bash history as the code?

## Option 1

use existing functionality
- ugly code using internals
- only 'multi': all commands need to pass SCT

```py
commands = get_bash_history()
for command in commands:
    Ex(Ex()._state.to_child(student_code=command)).has_code()
```

## Option 2

new general (dangerous?) ``update`` SCT to override arbitrary State attributes
alternative: update `override()` SCT with optional `student_code` argument (safer)

```py
def update(state, **kwargs):
    return state.to_child(**kwargs)
```

### Use cases:

- equivalent of `Ex().has_command(commands, pattern)`:
```py
Ex().check_or(*[update(student_code=command).has_code(pattern) for command in commands])
```


- shellwhat parsing of bash history:
```py
Ex().check_or(
    *[
        update(student_code=command)
        .check_node(node_name)
        .check_edge(edge_name)
        .has_equal_ast()
        for command in commands
    ]
)
```
"""

"""
SCT hooks?
let backend call an init function on exercise start
+ xwhat calls russian_protowhat init / register init handlers
"""



from copy import copy
from jinja2 import Template

from russian_protowhat.selectors import DispatcherInterface
from russian_protowhat.Feedback import Feedback, InstructorError
from russian_protowhat.Test import Fail, Test, TestFail
from russian_protowhat.utils import _debug


class DummyDispatcher(DispatcherInterface):
    def __init__(self):
        class ParseError(Exception):
            pass

        self.ParseError = ParseError

    def find(self, name, node, *args, **kwargs):
        return []

    def select(self, path: str, node):
        return None

    def parse(self, code):
        return self.ParseError()

    def describe(self, *args, **kwargs):
        return "code"


class State:
    def __init__(
        self,
        student_code,
        solution_code,
        pre_exercise_code,
        student_conn,
        solution_conn,
        student_result,
        solution_result,
        reporter,
        force_diagnose=False,
        highlighting_disabled=False,
        messages=None,
        creator=None,
        solution_ast=None,
        student_ast=None,
        ast_dispatcher=None,
    ):
        args = locals().copy()
        self.params = list()

        for k, v in args.items():
            if k != "self":
                self.params.append(k)
                setattr(self, k, v)

        self.messages = messages if messages else []

        if ast_dispatcher is None:
            self.ast_dispatcher = self.get_dispatcher()

        # Parse solution and student code
        # if possible, not done yet and wanted (ast arguments not False)
        if isinstance(self.solution_code, str) and self.solution_ast is None:
            self.solution_ast = self.parse(self.solution_code, test=False)
        if isinstance(self.student_code, str) and self.student_ast is None:
            self.student_ast = self.parse(self.student_code)

    def parse(self, text, test=True):
        result = None
        if self.ast_dispatcher:
            try:
                result = self.ast_dispatcher.parse(text)
            except self.ast_dispatcher.ParseError as e:
                if test:
                    self.report(e.message)
                else:
                    raise InstructorError(
                        "Something went wrong when parsing PEC or solution code: %s"
                        % str(e)
                    )

        return result

    def get_dispatcher(self):
        return DummyDispatcher()

    @property
    def parent_state(self):
        if self.creator is not None:
            return self.creator["args"]["state"]

    @property
    def is_root(self):
        return self.parent_state is None

    @property
    def state_history(self):
        history = [self]
        while history[-1].parent_state is not None:
            history.append(history[-1].parent_state)

        return list(reversed(history))

    def get_ast_path(self):
        rev_checks = filter(
            lambda x: x.creator is not None
            and x.creator["type"] in ["check_edge", "check_node"],
            reversed(self.state_history),
        )
        try:
            last = next(rev_checks)
            if last.creator["type"] == "check_node":
                # final check was for a node
                return self.ast_dispatcher.describe(
                    last.student_ast,
                    index=last.creator["args"]["index"],
                    msg="{index}{node_name}",
                )
            else:
                node = next(rev_checks)
                if node.creator["type"] == "check_node":
                    # checked for node, then for target, so can give rich description
                    return self.ast_dispatcher.describe(
                        node.student_ast,
                        field=last.creator["args"]["name"],
                        index=last.creator["args"]["index"],
                        msg="{index}{field_name} of the {node_name}",
                    )

        except StopIteration:
            return self.ast_dispatcher.describe(self.student_ast, "{node_name}")

    def report(self, feedback: str):
        test_feedback = Feedback(feedback, self)
        if test_feedback.highlight is None and self is not getattr(
            self, "root_state", None
        ):
            test_feedback.highlight = self.student_ast
        test = Fail(test_feedback)

        return self.do_test(test)

    def do_test(self, test: Test):
        result, feedback = self.reporter.do_test(test)
        if result is False:
            if getattr(self, "debug", False):
                setattr(self, "debug", False)  # prevent loop
                _debug(self)
            raise TestFail(feedback, self.reporter.build_failed_payload(feedback))
        return result, feedback

    def do_tests(self, tests):
        return [self.do_test(test) for test in tests]

    def to_child(self, append_message="", **kwargs):
        """Basic implementation of returning a child state"""

        bad_pars = set(kwargs) - set(self.params)
        if bad_pars:
            raise ValueError("Invalid init params for State: %s" % ", ".join(bad_pars))

        child = copy(self)
        for k, v in kwargs.items():
            setattr(child, k, v)

        # append messages
        if not isinstance(append_message, dict):
            append_message = {"msg": append_message, "kwargs": {}}
        child.messages = [*self.messages, append_message]

        return child

    def build_message(self, tail_msg="", fmt_kwargs=None, append=True):
        if not fmt_kwargs:
            fmt_kwargs = {}
        out_list = []
        # add trailing message to msg list
        msgs = self.messages[:] + [{"msg": tail_msg, "kwargs": fmt_kwargs}]

        # format messages in list, by iterating over previous, current, and next message
        for prev_d, d, next_d in zip([{}, *msgs[:-1]], msgs, [*msgs[1:], {}]):
            tmp_kwargs = {
                "parent": prev_d.get("kwargs"),
                "child": next_d.get("kwargs"),
                "this": d["kwargs"],
                **d["kwargs"],
            }
            # don't bother appending if there is no message
            if not d or not d["msg"]:
                continue
            # TODO: rendering is slow in tests (40% of test time)
            out = Template(d["msg"].replace("__JINJA__:", "")).render(tmp_kwargs)
            out_list.append(out)

        # if highlighting info is available, don't put all expand messages
        if getattr(self, "highlight", None) and not self.highlighting_disabled:
            out_list = out_list[-3:]

        if append:
            return "".join(out_list)
        else:
            return out_list[-1]



from typing import Dict


class Feedback:
    def __init__(self, message, state=None):
        self.message = message
        self.highlight = None
        self.highlighting_disabled = False
        self.path = None
        if state is not None:
            self.highlight = getattr(state, "highlight", None)
            self.highlighting_disabled = state.highlighting_disabled
            if hasattr(state, "path"):
                self.path = state.path

    def _highlight_data(self) -> Dict[str, int]:
        highlight = {}
        if self.path:
            highlight["path"] = self.path

        if hasattr(self.highlight, "get_position"):
            position = self.highlight.get_position()
            if position:
                highlight.update(position)
        elif getattr(self.highlight, "first_token", None) and getattr(
            self.highlight, "last_token", None
        ):
            # used by pythonwhat
            # a plugin+register system would be better
            # if many different AST interfaces exist
            highlight.update(
                {
                    "line_start": self.highlight.first_token.start[0],
                    "column_start": self.highlight.first_token.start[1],
                    "line_end": self.highlight.last_token.end[0],
                    "column_end": self.highlight.last_token.end[1],
                }
            )

        return highlight

    def get_highlight_data(self) -> Dict[str, int]:
        result = None
        if self.highlight is not None and not self.highlighting_disabled:
            result = self._highlight_data()

        return result or {}

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, repr(vars(self)))


class InstructorError(Exception):
    pass
